"""
https://pytorch.org/docs/stable/distributed.html
https://pytorch.org/docs/stable/data.html#torch.utils.data.distributed.DistributedSampler
https://pytorch.org/docs/stable/distributed.html#launch-utility

Complete working example:
https://github.com/pytorch/examples/blob/master/imagenet/main.py

Wrapper around nn.DataParallel and nn.parallel.DistributedDataParallel
"""
import os
import numpy as np
import torch
import torch.nn as nn
import torch.utils.data
import omlet.utils as U
import omlet.distributed.primitives as _primitives
from omlet.data.loaders import *
from .utils import *

if nvidia_apex_available():
    import apex.parallel

if horovod_available():
    import horovod.torch as hvd


__all__ = ["WorkerContext"]

# https://discuss.pytorch.org/t/issue-with-multiprocessing-semaphore-tracking/22943/4
os.environ["PYTHONWARNINGS"] = "ignore:semaphore_tracker:UserWarning"


class WorkerContext:
    def __init__(
        self,
        local_rank: int,
        global_rank: int,
        world_size: int,
        gpu_ids: List[int],
        backend: str,
        comm_backend: str,
        url: str,
        port: Optional[int],
        group,
        group_ids: Optional[List[int]],
    ):
        self.local_rank = local_rank
        self.global_rank = global_rank
        self.world_size = world_size
        self.gpu_ids = gpu_ids
        assert gpu_ids, "must specify at least one GPU ID"
        assert self.local_rank < len(gpu_ids), "internal error"
        self.backend = backend
        if backend == "single":
            assert len(self.gpu_ids) == 1, "internal error"
        self.comm_backend = comm_backend
        self.url = url
        self.port = port
        self.group = group
        self.group_ids = group_ids

    @property
    def gpu_id(self):
        if self.backend == "dp":
            raise ValueError("DP backend does not have a single GPU ID")
        return self.gpu_ids[self.local_rank]

    @property
    def device(self):
        return torch.device(self.gpu_id)

    def device_context(self):
        """
        Context manager
        """
        return torch.cuda.device(self.device)

    def is_default_group(self):
        return self.group == DEFAULT_GROUP

    def is_local_master(self):
        return self.local_rank == 0

    def is_global_master(self):
        return self.global_rank == 0

    @property
    def num_gpus(self) -> int:
        return len(self.gpu_ids)

    @property
    def num_procs(self) -> int:
        """
        world_size = sum of num_procs over all nodes
        """
        if self.backend in ["single", "dp"]:
            return 1
        else:
            return self.num_gpus

    def get_local_split_size(self, global_size):
        """
        Useful for computing batch size per GPU
        """
        if global_size % self.num_procs != 0:
            raise RuntimeError(
                f"global size {global_size} must divide number of processes {self.num_procs}"
            )
        return global_size // self.num_procs

    def data_parallel(
        self, model: nn.Module, sync_bn: bool = False, sync_bn_group=None, **kwargs
    ):
        """
        Args:
            model
            sync_bn: only available in pytorch >= 1.2 and Nvidia apex
            sync_bn_group: if None, defaults to syncBN over the entire `world`
            kwargs:
              - for Apex DDP: https://nvidia.github.io/apex/parallel.html#module-apex.parallel
        """
        backend = self.backend
        if sync_bn_group is None:
            sync_bn_group = DEFAULT_GROUP

        if backend in ["ddp", "ray"]:
            device = self.device
            if sync_bn:
                assert torch_syncbn_available(), (
                    "must use pytorch version >= 1.2 for SyncBN. "
                    "Otherwise you can use Nvidia apex syncBN."
                )
                model = nn.SyncBatchNorm.convert_sync_batchnorm(model, sync_bn_group)
            model.cuda(device)
            return nn.parallel.DistributedDataParallel(
                model, device_ids=[device], **kwargs
            )

        elif backend == "apex":
            from apex.parallel import (
                DistributedDataParallel as ApexDistributedDataParallel,
            )
            from apex.parallel import convert_syncbn_model as apex_convert_syncbn_model

            if sync_bn:
                model = apex_convert_syncbn_model(model, process_group=sync_bn_group)
            model.cuda(self.device)
            # defaults to True unless explicitly specified
            delay_allreduce = kwargs.pop("delay_allreduce", True)
            return ApexDistributedDataParallel(
                model, delay_allreduce=delay_allreduce, **kwargs
            )

        elif backend == "horovod":
            # TODO: horovod also wraps optimizer
            model.cuda(self.device)
            hvd.broadcast_parameters(model.state_dict(), root_rank=0)
            return model

        elif backend == "dp":
            return nn.DataParallel(model, device_ids=self.gpu_ids, **kwargs).cuda()

        elif backend == "single":
            return model.cuda(self.device)

        else:
            raise NotImplementedError(f"backend {backend} not supported")

    def unwrap(self, model):
        """
        Retrieve the model wrapped in DataParallel
        """
        if self.backend in ["single", "horovod"]:
            return model
        else:
            assert hasattr(model, "module"), (
                f"internal error: backend {self.backend}-wrapped model "
                f"has no `.module` attribute"
            )
            return model.module

    def get_data_loader(
        self,
        dataset,
        global_batch_size,
        num_workers,
        stage,
        epoch=None,
        keep_on_cpu=False,
        **loader_kwargs,
    ):
        # TODO
        raise NotImplementedError("TODO: port from omlet.distributed.legacy.manager")

    def set_seed(self, seed=111, same_worker_seed=True, deterministic=False):
        if not same_worker_seed and seed is not None:
            seed = seed + self.local_rank
        U.set_seed_everywhere(seed, deterministic=deterministic)

    def _torch_dist_enabled(self):
        return self.backend not in ["dp", "single"]

    def reduce(
        self,
        tensor,
        op="sum",
        broadcast: bool = True,
        inplace: bool = True,
        group=None,
    ) -> torch.Tensor:
        if self._torch_dist_enabled():
            return _primitives.reduce(
                tensor, op=op, broadcast=broadcast, inplace=inplace, group=group
            )
        else:
            if inplace:
                return tensor
            else:
                return tensor.clone().detach()

    def all_reduce_(self, tensor, op="sum", group=None):
        return self.reduce(tensor, op=op, group=group, broadcast=True, inplace=True)

    def all_reduce(self, tensor, op="sum", group=None):
        return self.reduce(tensor, op=op, group=group, broadcast=True, inplace=False)

    def master_reduce_(self, tensor, op="sum", group=None):
        return self.reduce(tensor, op=op, group=group, broadcast=False, inplace=True)

    def master_reduce(self, tensor, op="sum", group=None):
        return self.reduce(tensor, op=op, group=group, broadcast=False, inplace=False)

    def sync_barrier(self):
        if self._torch_dist_enabled():
            torchdist.barrier()

    def reduce_scalar(
        self, scalar, *, op="sum", broadcast=True, storage=None, group=None, device=None
    ):
        if self._torch_dist_enabled():
            return _primitives.reduce_scalar(
                scalar,
                op=op,
                broadcast=broadcast,
                storage=storage,
                group=group,
                device=device,
            )
        else:
            return float(scalar)

    def reduce_scalars(
        self,
        scalars,
        *,
        op="sum",
        broadcast=True,
        storage=None,
        group=None,
        device=None,
    ):
        if self._torch_dist_enabled():
            return _primitives.reduce_scalars(
                scalars,
                op=op,
                broadcast=broadcast,
                storage=storage,
                group=group,
                device=device,
            )
        else:
            return scalars
