"""
https://pytorch.org/docs/stable/distributed.html
https://pytorch.org/docs/stable/data.html#torch.utils.data.distributed.DistributedSampler
https://pytorch.org/docs/stable/distributed.html#launch-utility

Complete working example:
https://github.com/pytorch/examples/blob/master/imagenet/main.py

Wrapper around nn.DataParallel and nn.parallel.DistributedDataParallel
"""
import collections
import shlex
from datetime import timedelta

import torch.distributed as torchdist
import torch.utils.data
import omlet.utils as U
from .context import WorkerContext
from .utils import *
from typing import Union, Optional, Callable
from omlet.utils import omlet_logger as _omlet_log

if horovod_available():
    import horovod.run
    import horovod.torch as hvd
if ray_available():
    import ray


__all__ = [
    "DistLauncher",
    "LocalDistLauncher",
    "MultiNodeLauncher",
    "MultiNodeReplicaLauncher",
    "SingleGPULauncher",
    "DPLauncher",
]

# https://discuss.pytorch.org/t/issue-with-multiprocessing-semaphore-tracking/22943/4
os.environ["PYTHONWARNINGS"] = "ignore:semaphore_tracker:UserWarning"


class DistLauncher:
    BACKENDS = ALL_DIST_BACKENDS

    def __init__(
        self,
        *,
        backend: str,
        gpus: Union[str, int, List[int], None],
        world_size: Optional[int] = None,
        global_rank_start: int = 0,
        url: str = "localhost",
        port: Union[str, int, None] = "auto",
        process_groups: Union[int, List[int], List[List[int]], str, None] = None,
    ):
        self.gpu_ids = parse_gpu_ids(gpus)
        if not self.gpu_ids:
            raise NotImplementedError(
                "CPU mode is not yet supported. You must specify at least one GPU."
            )
        if world_size is None:
            world_size = len(self.gpu_ids)  # assume single node
        else:
            # fmt:off
            assert world_size >= len(self.gpu_ids), (
                f"world_size must be at least the number of GPU IDs: {self.gpu_ids}"
            )
            # fmt:on
        self.world_size = world_size
        assert global_rank_start < world_size
        self.global_rank_start = global_rank_start
        self.process_groups = parse_process_groups(process_groups, self.world_size)
        self.backend, self.comm_backend = self._parse_backend(backend)
        if self.backend in ["single", "dp"]:
            self.url, self.port = "", None
        else:
            self.url, self.port = parse_init_method_url(url, port)

        self._backend_settings = {}
        self._shared_settings = {}
        self.reset_backend_settings()

    def _parse_backend(self, backend):
        backend, comm_backend = parse_dist_backend(backend)
        if backend == "single":
            assert self.num_gpus == 1, "single backend requires one and only one GPU."
        return backend, comm_backend

    @property
    def num_gpus(self) -> int:
        return len(self.gpu_ids)

    @property
    def num_procs(self) -> int:
        if self.backend in ["single", "dp"]:
            return 1
        else:
            return self.num_gpus

    def backend_settings(
        self,
        *,
        nccl_timeout_s: Optional[int] = None,
        mpirun_bin_dir: Optional[str] = None,
        **kwargs,
    ):
        """
        For each backend, kwargs will be passed to the backend initialization or run
            method at launch() or launch_async()
            Note that some kwargs are filled by DistLauncher and cannot be overridden.
            The prefilled kwargs are indicated below before "**kwargs"

        ddp:
            torch.multiprocessing.spawn(fn, nprocs, args, join, **kwargs)

        ray:
            ray.init(**kwargs)
            nccl_timeout_s: for DDP timeout in seconds

        horovod:
            horovod.run.run(
                fn, kwargs, np, use_gloo, use_mpi, hosts, hostfile, ssh_port, **kwargs
            )

        Common settings:
        - nccl_timeout_s: for DDP timeout in seconds
            applies to all DDP-based backends: ddp:*, ray
        - mpirun_bin_dir (None): directory of the mpirun executable, if unspecified
            defaults to the system-wide mpirun.
            applies to all MPI-based backends: horovod:mpi, ddp:mpi
        """
        self._shared_settings.update(
            dict(nccl_timeout_s=nccl_timeout_s, mpirun_bin_dir=mpirun_bin_dir)
        )
        self._backend_settings.update(kwargs)
        return self

    def reset_backend_settings(self):
        self._shared_settings = {
            "nccl_timeout_s": None,
            "mpirun_bin_dir": None,
        }
        self._backend_settings = {}

    def launch(self, worker_fn: Callable, *worker_args, **worker_kwargs):
        self._launch(worker_fn, worker_args, worker_kwargs, async_=False)

    def launch_async(self, worker_fn: Callable, *worker_args, **worker_kwargs):
        self._launch(worker_fn, worker_args, worker_kwargs, async_=True)

    def _launch(self, worker_fn, worker_args, worker_kwargs, async_: bool):
        """
        Special kwargs:
            ray_init_kwargs: dict kwargs to pass to ray.init()
        """
        if self.backend in ["ddp", "apex"]:
            self._launch_ddp(worker_fn, worker_args, worker_kwargs, async_)
        elif self.backend in ["single", "dp"]:
            self._launch_single_proc(worker_fn, worker_args, worker_kwargs, async_)
        elif self.backend == "ray":
            self._launch_ray(worker_fn, worker_args, worker_kwargs, async_)
        elif self.backend == "horovod":
            self._launch_horovod(worker_fn, worker_args, worker_kwargs, async_)
        else:
            raise NotImplementedError(f"INTERNAL: unrecognized backend {self.backend}")

    class _DummyRayActor:
        def __init__(self, launcher, cuda_visible_devices):
            # reference back to the launcher itself
            self.launcher = launcher
            # pass master's env and override Ray's CUDA_VISIBLE_DEVICES selection
            if cuda_visible_devices is None:
                os.environ.pop("CUDA_VISIBLE_DEVICES")
            else:
                os.environ["CUDA_VISIBLE_DEVICES"] = cuda_visible_devices

        def run(self, local_rank, worker_fn, worker_args, worker_kwargs):
            self.launcher._worker_wrapper(
                local_rank, worker_fn, worker_args, worker_kwargs
            )

    def _launch_ray(self, worker_fn, worker_args, worker_kwargs, async_):
        # TODO: ray.init needs to support multinode URL from master node
        if not ray.is_initialized():
            ray.init(**self._backend_settings)
        ip_addr = ray.services.get_node_ip_address()
        # override URL to redirect to Ray's address
        self.url = f"tcp://{ip_addr}:{self.port}"
        rids = []
        DummyActor = ray.remote(num_gpus=1)(DistLauncher._DummyRayActor)
        cuda_visible_devices = os.environ.get("CUDA_VISIBLE_DEVICES", None)
        for local_rank in range(self.num_procs):
            runner = DummyActor.remote(self, cuda_visible_devices)
            rid = runner.run.remote(local_rank, worker_fn, worker_args, worker_kwargs)
            rids.append(rid)
        if async_:
            return
        else:
            ray.get(rids)

    def _launch_ddp(self, worker_fn, worker_args, worker_kwargs, async_):
        torch.multiprocessing.spawn(
            self._worker_wrapper,
            nprocs=self.num_procs,
            args=(worker_fn, worker_args, worker_kwargs),
            join=not async_,
            **self._backend_settings,
        )

    def _launch_single_proc(self, worker_fn, worker_args, worker_kwargs, async_):
        """
        This will always block regardless of async_
        """
        if async_:
            _omlet_log.warning(
                f"Backend {self.backend} does not support async and "
                f"will always block until finish"
            )
        self._worker_wrapper(0, worker_fn, worker_args, worker_kwargs)

    def _launch_horovod(self, worker_fn, worker_args, worker_kwargs, async_):
        # pass a dummy local_rank because Horovod sets its own local_rank
        if async_:
            _omlet_log.warning(
                f"Horovod does not support async and will always block until finish"
            )
        horovod_kwargs = self._backend_settings
        # specify an alternative mpirun executable path
        mpirun_bin_dir = self._shared_settings["mpirun_bin_dir"]
        if mpirun_bin_dir:
            mpirun_bin_dir = U.f_expand(mpirun_bin_dir)
            assert os.path.isdir(mpirun_bin_dir)
            os.environ["PATH"] = f'{shlex.quote(mpirun_bin_dir)}:{os.environ["PATH"]}'
            # os.system('mpirun --version')
        # WARNING: you need -oversubscribe for MPI > 3.0
        # https://github.com/horovod/horovod/issues/1691
        mpi_args = horovod_kwargs.get("mpi_args", "")
        if "-oversubscribe" not in mpi_args:
            mpi_args += " -oversubscribe"
        horovod_kwargs["mpi_args"] = mpi_args

        horovod.run.run(
            self._worker_wrapper,
            kwargs=dict(
                local_rank=None,
                worker_fn=worker_fn,
                worker_args=worker_args,
                worker_kwargs=worker_kwargs,
            ),
            np=self.num_procs,
            use_gloo=self.comm_backend == "gloo",
            use_mpi=self.comm_backend == "mpi",
            hosts=None,  # TODO: multinode, hack self.url to support this
            hostfile=None,  # TODO: multinode, hack self.url "file://"
            ssh_port=None,  # TODO: multinode, hack self.port
            **horovod_kwargs,
        )

    def _worker_wrapper(self, local_rank, worker_fn, worker_args, worker_kwargs):
        global_rank = 0
        group = None
        group_ids = None

        if self.backend in ["ddp", "apex", "ray"]:
            global_rank = self.global_rank_start + local_rank
            # pytorch 1.5: torch.distributed.constants.default_pg_timeout
            timeout = timedelta(minutes=30)
            if self._shared_settings["nccl_timeout_s"]:
                timeout = timedelta(seconds=self._shared_settings["nccl_timeout_s"])
                if self.comm_backend == "nccl":
                    # see torch.distributed.init_process_group doc
                    os.environ["NCCL_BLOCKING_WAIT"] = "1"

            torchdist.init_process_group(
                backend=self.comm_backend,
                init_method=self.url,
                world_size=self.world_size,
                rank=global_rank,
                timeout=timeout,
            )
            # all procs must reach `new_group()` call even when they don't belong
            if self.process_groups:
                for gids in self.process_groups:
                    g = torchdist.new_group(gids, backend=self.comm_backend)
                    # if multiple groups share the same process, we only assign
                    # the process to the first group in which it appears
                    if global_rank in gids and group is None:
                        group = g
                        group_ids = gids

        elif self.backend == "horovod":
            assert local_rank is None, "INTENRAL ERROR: horovod sets its own local_rank"
            # TODO horovod does not yet support groups, but hvd.init(group) can be done
            # TODO: horovod doesn't support multinode, requires SSH
            hvd.init()
            local_rank = hvd.local_rank()
            global_rank = hvd.rank()
            assert self.world_size == hvd.size(), (
                f"INTERNAL ERROR: world_size {self.world_size} "
                f"should be equal to horovod global size {hvd.size()}"
            )
            assert self.num_procs == hvd.local_size(), (
                f"INTERNAL ERROR: number of local procs {self.num_procs} "
                f"should be equal to horovod local size {hvd.local_size()}"
            )

        if group is None:
            group = DEFAULT_GROUP

        U.set_os_envs(
            {
                "OMLET_DIST_INITIALIZED": 1,
                "OMLET_LOCAL_RANK": local_rank,
                "OMLET_GLOBAL_RANK": global_rank,
                "OMLET_WORLD_SIZE": self.world_size,
                "OMLET_MASTER_URL": self.url,
                "OMLET_MASTER_PORT": "" if self.port is None else self.port,
                "OMLET_BACKEND": self.backend,
                "OMLET_COMM_BACKEND": self.comm_backend,
            }
        )

        ctx = WorkerContext(
            local_rank=local_rank,
            global_rank=global_rank,
            world_size=self.world_size,
            gpu_ids=self.gpu_ids,
            backend=self.backend,
            comm_backend=self.comm_backend,
            url=self.url,
            port=self.port,
            group=group,
            group_ids=group_ids,
        )
        with ctx.device_context():
            worker_fn(ctx, *worker_args, **worker_kwargs)


class LocalDistLauncher(DistLauncher):
    def __init__(
        self,
        gpus: Union[str, int, List[int], None],
        backend: str = "ddp:nccl",
        port: Union[str, int, None] = "auto",
        process_groups: Union[int, List[int], List[List[int]], str, None] = None,
    ):
        super().__init__(
            backend=backend,
            gpus=gpus,
            world_size=None,
            url="localhost",
            port=port,
            process_groups=process_groups,
        )


class MultiNodeLauncher(DistLauncher):
    """
    Each node has exactly the same number of GPUs
    """

    def __init__(
        self,
        backend: str,
        node_id: int,
        node_gpus: List[Union[str, int, List[int]]],
        url: str = "localhost",
        port: Union[str, int, None] = "auto",
        process_groups: Union[int, List[int], List[List[int]], str, None] = None,
    ):
        """
        Args:
            node_id: current node index
            node_gpus: list of GPU specs, each one has the same format
                as DistLauncher `gpus` arg
            url: master node's URL
            port: master node's port
        """
        assert backend not in ["single", "dp"]
        node_id = int(node_id)

        assert isinstance(node_gpus, (list, tuple))
        assert (
            0 <= node_id < len(node_gpus)
        ), f"node_id {node_id} must be smaller than number of nodes: len{node_gpus}"
        # e.g. [[0,1,2], [4,5,6,7], [0,1,2,3,4]]
        parsed_node_gpus = []
        for gpus in node_gpus:
            assert gpus not in [-1, "-1"], (
                f'{self.__class__.__name__} does not support "-1" as the GPU spec, '
                f"you must specify the exact GPU IDs or number of GPUs on each node."
            )
            g = parse_gpu_ids(gpus)
            assert g, f"must have at least one GPU: {gpus} in node_gpus {node_gpus}"
            parsed_node_gpus.append(g)

        world_size = sum(len(gpus) for gpus in parsed_node_gpus)
        global_rank_start = sum(len(gpus) for gpus in parsed_node_gpus[:node_id])

        super().__init__(
            backend=backend,
            gpus=parsed_node_gpus[node_id],
            world_size=world_size,
            global_rank_start=global_rank_start,
            url=url,
            port=port,
            process_groups=process_groups,
        )


class MultiNodeReplicaLauncher(DistLauncher):
    """
    Each node has exactly the same number of GPUs
    """

    def __init__(
        self,
        backend: str,
        node_id: int,
        num_nodes: int,
        num_gpus_per_node: int,
        url: str = "localhost",
        port: Union[str, int, None] = "auto",
        process_groups: Union[int, List[int], List[List[int]], str, None] = None,
    ):
        """
        Args:
            node_id: current node index
            url: master node's URL
            port: master node's port
        """
        assert backend not in ["single", "dp"]
        node_id = int(node_id)
        num_nodes = int(num_nodes)
        assert num_nodes > 0
        assert 0 <= node_id < num_nodes
        num_gpus_per_node = int(num_gpus_per_node)

        if num_gpus_per_node <= 0:
            num_gpus_per_node = torch.cuda.device_count()

        super().__init__(
            backend=backend,
            gpus=num_gpus_per_node,
            world_size=num_nodes * num_gpus_per_node,
            global_rank_start=node_id * num_gpus_per_node,
            url=url,
            port=port,
            process_groups=process_groups,
        )


class SingleGPULauncher(DistLauncher):
    def __init__(
        self, gpu_id: Union[str, int],
    ):
        super().__init__(
            backend="single",
            gpus=gpu_id,
            world_size=1,
            url="",
            port=None,
            process_groups=None,
        )


class DPLauncher(DistLauncher):
    def __init__(
        self, gpus: Union[str, int, List[int], None],
    ):
        super().__init__(
            backend="dp",
            gpus=gpus,
            world_size=1,
            url="",
            port=None,
            process_groups=None,
        )
