import os
import torch
import torch.distributed as torchdist
import torch.utils.data
from typing import Union, Optional
from .utils import *

if horovod_available():
    import horovod.torch as hvd


def reduce(
    tensor, op="sum", broadcast=True, inplace: bool = True, group=None
) -> torch.Tensor:
    backend = get_backend()
    if backend in ["ddp", "ray", "apex"]:
        method = _torch_reduce
    elif backend == "horovod":
        method = _horovod_reduce
    else:
        raise ValueError(f"Backend {backend} does not support reduce operation.")

    return method(tensor, op=op, broadcast=broadcast, inplace=inplace, group=group)


def _horovod_reduce(
    tensor, op="sum", broadcast=True, inplace: bool = True, group=None
) -> torch.Tensor:
    assert broadcast, "Horovod only supports broadcast=True"
    _HOROVOD_REDUCE_OPS = {
        "sum": hvd.Sum,
        "mean": hvd.Average,
    }
    op_name = op.lower()
    assert op_name in _HOROVOD_REDUCE_OPS, (
        f"invalid horovod reduce op: {op}. "
        f"All available ops: {list(_HOROVOD_REDUCE_OPS.keys())}"
    )
    op = _HOROVOD_REDUCE_OPS[op_name]

    assert torch.is_tensor(tensor)
    if inplace:
        return hvd.allreduce_(tensor, op=op)
    else:
        return hvd.allreduce(tensor, op=op)


_TORCH_REDUCE_OPS = {
    "sum": torchdist.ReduceOp.SUM,
    "mean": torchdist.ReduceOp.SUM,  # no buildin for mean, we handle it ourselves
    "product": torchdist.ReduceOp.PRODUCT,
    "max": torchdist.ReduceOp.MAX,
    "min": torchdist.ReduceOp.MIN,
    "and": torchdist.ReduceOp.BAND,
    "or": torchdist.ReduceOp.BOR,
    "xor": torchdist.ReduceOp.BXOR,
}


def _torch_reduce(
    tensor, op: str = "sum", broadcast: bool = True, inplace: bool = True, group=None
) -> torch.Tensor:
    """
    Args:
        tensor
        broadcast: True for all-reduce, False for master-only
        out:
            - if the same tensor as input, will mutate input
            - None: create a new tensor and leave the input unchanged
    """
    op_name = op.lower()
    assert op_name in _TORCH_REDUCE_OPS, (
        f"invalid torch reduce op: {op}. "
        f"All available ops: {list(_TORCH_REDUCE_OPS.keys())}"
    )
    op = _TORCH_REDUCE_OPS[op_name]

    assert torch.is_tensor(tensor)
    if inplace:
        out = tensor
    else:
        out = tensor.clone().detach()
    group = group or DEFAULT_GROUP
    if broadcast:
        torchdist.all_reduce(out, op=op, group=group)
    else:
        backend = get_comm_backend()
        if backend == "gloo" and out.is_cuda:
            raise RuntimeError("gloo backend does not support broadcast=False on GPU.")
        torchdist.reduce(out, dst=0, op=op, group=group)
    if op_name == "mean" and (broadcast or is_global_master(group)):
        out.div_(1.0 * get_world_size(group))
    return out


def all_reduce_(tensor, op="sum", group=None):
    return reduce(tensor, op=op, group=group, broadcast=True, inplace=True)


def all_reduce(tensor, op="sum", group=None):
    return reduce(tensor, op=op, group=group, broadcast=True, inplace=False)


def master_reduce_(tensor, op="sum", group=None):
    return reduce(tensor, op=op, group=group, broadcast=False, inplace=True)


def master_reduce(tensor, op="sum", group=None):
    return reduce(tensor, op=op, group=group, broadcast=False, inplace=False)


def reduce_scalar(
    scalar, op="sum", *, broadcast=True, storage=None, group=None, device=None
):
    """
    Args:
        scalar: single python float
        broadcast: True to use all_reduce, else master_reduce
        storage: if you call this method repeatedly, consider allocating
            a persistent CUDA tensor in advance to provide temporary storage
            if None, allocates new CUDA memory every time
        device: if None, defaults to torch.cuda.current_device()
    Returns:
        python float
    """
    scalar = float(scalar)
    is_nccl = get_comm_backend() == "nccl"
    if storage is None:
        s = torch.tensor([scalar], dtype=torch.float32)
        if is_nccl:
            if device is None:
                device = torch.cuda.current_device()
            s = s.cuda(device)
    else:
        assert torch.is_tensor(storage)
        assert storage.numel() > 0, "storage must have at least 1 element"
        if is_nccl:
            assert storage.is_cuda
        storage[0] = scalar
        s = storage
    reduce(s, op=op, group=group, broadcast=broadcast, inplace=True)
    return float(s[0])


def reduce_scalars(
    scalars, op="sum", *, broadcast=True, storage=None, group=None, device=None
):
    """
    Args:
        scalars: list, tuple or dict of python floats
        broadcast: True to use all_reduce, else master_reduce
        storage: if you call this method repeatedly, consider allocating
            a persistent CUDA tensor in advance to provide temporary storage
            if None, allocates new CUDA memory every time
        device: if None, defaults to torch.cuda.current_device()
    Returns:
        list, tuple or dict of reduced floats
    """
    if not isinstance(scalars, (tuple, list, dict)):
        # singleton falls back to reduce_scalar()
        return reduce_scalar(
            scalars,
            op=op,
            broadcast=broadcast,
            storage=storage,
            group=group,
            device=device,
        )

    is_nccl = get_comm_backend() == "nccl"
    idx_map = None
    if isinstance(scalars, dict):
        values = []
        idx_map = {}
        for i, (k, v) in enumerate(scalars.items()):
            idx_map[k] = i
            values.append(float(v))
    else:
        values = [float(v) for v in scalars]
    numel = len(values)

    values = torch.tensor(values, dtype=torch.float32)
    if storage is None:
        if is_nccl:
            if device is None:
                device = torch.cuda.current_device()
            values = values.cuda(device)
    else:
        assert torch.is_tensor(storage)
        assert storage.numel() >= numel, "storage not enough"
        if is_nccl:
            assert storage.is_cuda
        storage[:numel] = values
        values = storage
    reduce(values, op=op, group=group, broadcast=broadcast, inplace=True)

    if isinstance(scalars, dict):
        return {k: float(values[idx_map[k]]) for k in idx_map}
    else:
        return type(scalars)(float(v) for v in values[:numel])
