import functools
import os
import re
import socket
import torch
import torch.distributed as torchdist
from omlet.utils.functional_utils import meta_decorator
from typing import List, Union, Optional, Any

try:
    import apex.parallel

    _NVIDIA_APEX_AVAILABLE = True
except ImportError:
    _NVIDIA_APEX_AVAILABLE = False


try:
    import ray

    _RAY_AVAILABLE = True
except ImportError:
    _RAY_AVAILABLE = False


try:
    # install command:
    # HOROVOD_GPU_OPERATIONS=NCCL pip install horovod
    import horovod.torch as hvd

    _HOROVOD_AVAILABLE = True
except ImportError:
    _HOROVOD_AVAILABLE = False


try:
    import mpi4py

    _MPI_AVAILABLE = True
except ImportError:
    _MPI_AVAILABLE = False


ALL_DIST_BACKENDS = [
    "single",
    "dp",
    "ddp",  # defaults to ddp:nccl
    "ddp:nccl",
    "ddp:gloo",
    "ddp:mpi",
    "apex",  # requires Nvidia apex installation
    "ray",
    "horovod",  # defaults to horovod:mpi
    "horovod:mpi",
    "horovod:gloo",
]

DEFAULT_GROUP = torchdist.group.WORLD


def nvidia_apex_available(check=False):
    if check and not _NVIDIA_APEX_AVAILABLE:
        raise ImportError("Nvidia apex is not installed.")
    return _NVIDIA_APEX_AVAILABLE


def ray_available(check=False):
    if check and not _RAY_AVAILABLE:
        raise ImportError("Ray is not installed.")
    return _RAY_AVAILABLE


def horovod_available(check=False):
    if check and not _HOROVOD_AVAILABLE:
        raise ImportError("Horovod is not installed.")
    return _HOROVOD_AVAILABLE


def mpi_available(check=False):
    if check and not _MPI_AVAILABLE:
        raise ImportError("mpi4py is not installed")
    return _MPI_AVAILABLE


def torch_syncbn_available():
    """
    PyTorch >= 1.2 supports SyncBN
    PyTorch 1.1 supports too, but there is a process_group bug:

    It didn't pass in process_group to children module recursively
    https://github.com/pytorch/pytorch/pull/19240
    """
    from pkg_resources import parse_version

    return parse_version(torch.__version__) >= parse_version("1.2.0")


def parse_dist_backend(backend: str):
    """
    Returns:
        (backend, comm_backend)
    """
    assert backend in ALL_DIST_BACKENDS, f"Unsupported backend: {backend}"
    comm_backend = "nccl"  # default GPU communication engine

    if backend in ["single", "dp"]:
        comm_backend = ""

    if backend == "horovod":
        # default communicator for horovod is MPI
        # NCCL support must be built at horovod installation (compilation) time
        comm_backend = "mpi"

    if ":" in backend:
        backend, comm_backend = backend.split(":")

    if comm_backend == "mpi":
        mpi_available(check=True)

    if backend == "apex":
        nvidia_apex_available(check=True)
    elif backend == "ray":
        ray_available(check=True)
    elif backend == "horovod":
        horovod_available(check=True)

    return backend, comm_backend


# ========================================================
# ================ Read OMLET os env vars ================
# ========================================================


def is_omlet_dist_initialized():
    return os.environ.get("OMLET_DIST_INITIALIZED") == "1"


def get_backend():
    if is_omlet_dist_initialized():
        return os.environ["OMLET_BACKEND"]
    elif torchdist.is_initialized():
        return "ddp"
    else:
        raise RuntimeError(
            f"distributed pkg uninitialized, cannot get distributed backend"
        )


def get_comm_backend():
    if is_omlet_dist_initialized():
        return os.environ["OMLET_COMM_BACKEND"]
    elif torchdist.is_initialized():
        return torchdist.get_backend()
    else:
        raise RuntimeError(
            f"distributed pkg uninitialized, cannot get communication backend"
        )


def get_local_rank():
    """
    get_local_rank does not support group intentionally.
    group only makes sense in the global context
    """
    if is_omlet_dist_initialized():
        return int(os.environ["OMLET_LOCAL_RANK"])
    else:
        return 0


def is_local_master():
    return get_local_rank() == 0


def _check_support_pg(group):
    backends = ["ddp", "ray", "apex"]
    assert get_backend() in backends, f"Only {backends} backends support process groups"
    if group is not None:
        raise NotImplementedError("TODO: support process group")


def get_global_rank(group=None):
    if is_omlet_dist_initialized():
        _check_support_pg(group)
        return int(os.environ["OMLET_GLOBAL_RANK"])
    elif torchdist.is_initialized():
        return torchdist.get_rank(group or DEFAULT_GROUP)
    else:
        return 0


def is_global_master(group=None):
    return get_global_rank(group) == 0


@meta_decorator
def global_master_only(fn, default_return: Any = None):
    """
    Use as a decorator for methods that should only execute on the global master process
    """

    @functools.wraps(fn)
    def wrapped_fn(*args, **kwargs):
        if is_global_master():
            return fn(*args, **kwargs)
        else:
            return default_return

    return wrapped_fn


@meta_decorator
def local_master_only(fn, default_return: Any = None):
    """
    Use as a decorator for methods that should only execute on the global master process
    """

    @functools.wraps(fn)
    def wrapped_fn(*args, **kwargs):
        if is_local_master():
            return fn(*args, **kwargs)
        else:
            return default_return

    return wrapped_fn


def get_world_size(group=None):
    if is_omlet_dist_initialized():
        _check_support_pg(group)
        return int(os.environ["OMLET_WORLD_SIZE"])
    elif torchdist.is_initialized():
        return torchdist.get_world_size(group or DEFAULT_GROUP)
    else:
        return 0


# ========================================================
# =================== GPUs ====================
# ========================================================


def current_device():
    return torch.cuda.current_device()


def optimize_cudnn():
    import torch.backends.cudnn as _cudnn

    _cudnn.benchmark = True
    _cudnn.deterministic = False


def random_free_tcp_port() -> int:
    tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    tcp.bind(("", 0))
    addr, port = tcp.getsockname()
    tcp.close()
    return port


def get_all_available_gpus():
    """
    :return: a list of all available gpus
    """
    return list(range(torch.cuda.device_count()))


def parse_gpu_ids(gpus: Union[int, List[int], str, None]) -> List[int]:
    """
    Spec: https://pytorch-lightning.readthedocs.io/en/latest/multi_gpu.html#gpu-device-selection

    Examples:
      - None: CPU (returns empty list)
      - "cpu": CPU
      - 0: CPU
      - 3: first 3 GPUs
      - -1: all available GPUs
      - [0]: GPU 0
      - [1, 3]: GPUs 1 and 3
      - '0': GPU 0
      - '3': GPU 3
      - '1,3': GPUs 1 and 3
      - '-1': all available GPUs
    """
    if gpus is None:
        return []

    all_gpus = get_all_available_gpus()

    if isinstance(gpus, int):
        if gpus < 0:
            selected = all_gpus
        else:
            # if gpus == 0, return [] (CPU only)
            selected = list(range(gpus))
    elif isinstance(gpus, list):
        selected = gpus
    elif isinstance(gpus, str):
        if gpus == "cpu":
            return []
        if "," in gpus:
            selected = [int(g) for g in gpus.split(",")]
        else:
            gpus = int(gpus)
            if gpus < 0:
                selected = all_gpus
            else:
                selected = [gpus]
    else:
        raise ValueError(f"gpus arg {gpus} (type {type(gpus)}) is not supported")

    assert all(
        isinstance(g, int) and g >= 0 for g in selected
    ), "all GPU IDs must be positive ints"
    selected = set(selected)
    assert selected.issubset(set(all_gpus)), (
        f"GPU IDs {list(selected - set(all_gpus))} are not available. "
        f"All available GPUs: {all_gpus}"
    )
    return sorted(list(selected))


def parse_process_groups(
    pg: Union[int, List[int], List[List[int]], str, None], world_size: int
):
    """
    Examples:
      - None: use the default global process group
      - "": empty string, same as above
      - 4: partition `world_size` into 4 equal groups
      - "4": same as above
      - [4, 4, 2]: partition `world_size` into 3 groups,
        which have 4, 4, and 2 processes respectively.
      - "4;4;2": same as above
      - [[0,2,4], [1,5,6], [3,7]]: specify the process ranks in each process group,
            e.g. process 0, 2, 4 belong to the first group,
                process 1, 5, 6 belong to the second group,
                process 3, 7 belong to the third group.
      - "0,2,4;1,5,6;3,7": same as above
    """
    if pg is None or pg == "":
        # default process group
        return None
    elif isinstance(pg, int):
        assert (
            pg > 0 and world_size % pg == 0
        ), f"world_size {world_size} must divide number of process groups {pg}"
        pg_size = world_size // pg
        return [list(range(i * pg_size, (i + 1) * pg_size)) for i in range(pg)]
    elif isinstance(pg, (list, tuple)):
        if isinstance(pg[0], int):
            # 4, 4, 2 -> partition world size into 3 groups
            assert sum(pg) == world_size, (
                f"sum of number of processes in each group {pg} "
                f"must match world size {world_size}"
            )
            pg_spec = []
            s = 0
            for pg_size in pg:
                pg_spec.append(list(range(s, s + pg_size)))
                s += pg_size
            return pg_spec
        elif isinstance(pg[0], (list, tuple)):
            assert sum(map(len, pg)) == world_size, (
                f"sum of number of processes in each group {pg} "
                f"must match world size {world_size}"
            )
            # convert to nested list
            pg = [list(p) for p in pg]
            overlap = set()
            for p in pg:
                # fmt: off
                assert len(p) == len(set(p)), (
                    f"duplicate process within a process group: {p}"
                )
                # fmt: on
                assert not (set(p) & overlap), (
                    f"a process should only belong to a single process group, "
                    f"found duplicate processes: {set(p) & overlap}"
                )
                overlap.update(set(p))
            return pg

    elif isinstance(pg, str):
        if ";" in pg:
            pg = [[int(p) for p in p_sub.split(",")] for p_sub in pg.split(";")]
            return parse_process_groups(pg, world_size)
        elif "," in pg:
            pg = [int(p) for p in pg.split(",")]
            return parse_process_groups(pg, world_size)
        else:
            return parse_process_groups(int(pg), world_size)
    else:
        raise ValueError(f"Unsupported process group spec: {pg}, type {type(pg)}")


"""
https://www.oreilly.com/library/view/regular-expressions-cookbook/9781449327453/ch08s11.html

\A
[a-z][a-z0-9+\-.]*://               # Scheme
([a-z0-9\-._~%!$&'()*+,;=]+@)?      # User
([a-z0-9\-._~%]+                    # Named or IPv4 host
|\[[a-z0-9\-._~%!$&'()*+,;=:]+\])   # IPv6+ host
:(?P<port>[0-9]+)                   # Port number
"""

_extract_port_re = re.compile(
    r"""
\A
[a-z][a-z0-9+\-.]*://
([a-z0-9\-._~%!$&'()*+,;=]+@)?
([a-z0-9\-._~%]+
|\[[a-z0-9\-._~%!$&'()*+,;=:]+\])
:(?P<port>[0-9]+)
""".replace(
        "\n", ""
    )
)


def extract_port_from_url(url: str):
    match = _extract_port_re.match(url)
    if match is None:
        return None
    else:
        return int(match.group("port"))


def parse_init_method_url(url: str, port: Union[str, int, None] = "auto"):
    if url.startswith("env://") or url.startswith("file://"):
        return url, None
    else:
        if not url.startswith("tcp://"):
            url = "tcp://" + url
        extracted_port = extract_port_from_url(url)
        if extracted_port is None:
            # port not specified in URL, use automatic port
            if port == "auto":
                port = random_free_tcp_port()
            else:
                port = int(port)
            assert port > 0, f"port must be a positive int: {port}"
            return f"{url}:{port}", port
        else:
            port = extracted_port
            assert port > 0, f"port must be a positive int: {port}"
            return f"{url}", port


def get_cuda_visible_devices() -> List[int]:
    """
    parse CUDA_VISIBLE_DEVICES
    """
    if "CUDA_VISIBLE_DEVICES" not in os.environ:
        return []
    devices = os.environ["CUDA_VISIBLE_DEVICES"]
    return [int(g) for g in devices.strip(" ,").split(",")]
