"""
Track various metrics at the end of batch and/or epochs
"""
import functools
from typing import Optional, Union, Callable

import omlet.utils as U


@U.state_dict_class(["value", "total_size", "count", "history"])
@functools.total_ordering
class Metric:
    def __init__(
        self,
        name: str,
        avg_mode: str = "simple",
        ema_decay: Optional[float] = None,
        ema_nsamples: Optional[int] = None,
        maximize: Optional[bool] = True,
        keep_history: bool = False,
        formatter: Union[str, Callable[[str], str], None] = None,
        comparison_tol: float = 1e-12,
        # for annotation only, used in Session
        sync_workers: bool = False,
        scope: Union[int, str] = "epoch",
    ):
        """
        Args:
            avg_mode: "simple", "ema", or "none"
            ema_decay: None for simple cumulative averaging
                otherwise 0 <= ema_decay <= 1
                0 means you always return the last value as the EMA
                1 means you don't update and only return the first value as the EMA
            ema_nsamples: mutually exclusive with ema_decay
                ema_decay will be calculated as 1. - 2/(N+1)
                https://en.wikipedia.org/wiki/Moving_average#Relationship_between_SMA_and_EMA

            comparison_tol: difference below which will be considered equal
            scope: possible values:
                - an integer for reset every N batches
                - "epoch" or "lifetime"
            sync_workers: annotation ONLY, Session handles sync over distributed workers
        """
        self.name = name
        self.sync_workers = sync_workers
        assert isinstance(scope, int) and scope >= 1 or scope in ["epoch", "lifetime"]
        self.scope = scope
        assert maximize in [True, False, None]
        self.maximize = maximize
        if formatter:
            assert isinstance(formatter, str) or callable(formatter)
        self.formatter = formatter
        self.comparison_tol = comparison_tol
        self.value = self.total_size = self.count = None
        self.keep_history = keep_history
        self.lifetime_count = 0  # never resets to 0
        self.history = []
        self.ema_decay = None
        assert avg_mode in ["simple", "ema", "none"]
        self.avg_mode = avg_mode
        self.set_ema(ema_decay, ema_nsamples)
        if avg_mode == "simple":
            assert (
                self.ema_decay is None
            ), "simple averaging does not support EMA options: ema_decay and ema_nsamples"
        elif avg_mode == 'ema':
            assert (
                self.ema_decay is not None
            ), "must specify either ema_decay or ema_nsamples"
        self.clear()

    def clear(self):
        self.value = 0.0
        self.total_size = 0
        self.count = 0

    def new_epoch(self):
        """
        If scope is N batches or "epoch", it will get cleared
        """
        if self.scope != "lifetime":
            if self.keep_history:
                # history (value, lifetime_count)
                # do not duplicate save value
                if len(self.history) == 0 or self.history[-1][1] != self.lifetime_count:
                    self.history.append((self.value, self.lifetime_count))
            self.clear()

    def set_ema(self, decay=None, nsamples=None):
        assert (
            decay is None or nsamples is None
        ), "cannot specify both ema decay and nsamples, please specify only one."
        if decay is not None:
            assert 0 <= decay <= 1
        elif nsamples is not None:
            decay = 1 - 2.0 / (nsamples + 1)
        self.ema_decay = decay
        return self.ema_decay

    def update(self, value: float, batch_size: int = 1):
        """
        Args:
            value: mean value over the batch
            batch_size
        """
        if (
            isinstance(self.scope, int)
            and self.count > 0
            and self.count % self.scope == 0
        ):
            if self.keep_history:
                self.history.append((self.value, self.lifetime_count))
            self.clear()

        new_value = float(value)
        batch_size = int(batch_size)

        self.count += 1
        self.lifetime_count += 1
        self.total_size += batch_size
        if self.avg_mode == "none" or self.count == 1:
            self.value = new_value
            return

        if self.avg_mode == "simple":
            decay = 1 - batch_size / self.total_size
        elif self.avg_mode == "ema":
            decay = self.ema_decay ** batch_size
        else:
            raise NotImplementedError(f"Internal: avg_mode {self.avg_mode}")
        self.value = decay * self.value + (1 - decay) * new_value

    def state_dict(self):
        return {
            key: getattr(self, key)
            for key in ["value", "total_size", "count", "history"]
        }

    def load_state_dict(self):
        pass

    def __float__(self):
        return float(self.value)

    def __str__(self):
        if self.formatter is None:
            return str(self.value)
        elif isinstance(self.formatter, str):
            return self.formatter.format(self.value)
        elif callable(self.formatter):
            return self.formatter(self.value)
        else:
            raise NotImplementedError(f"Unrecognized formatter: {self.formatter}")

    def __repr__(self):
        return (
            f"{self.__class__.__name__}{{"
            f"value={self.value}, count={self.count}"
            f"}}"
        )

    def _check_comparable(self):
        if self.maximize is None:
            raise ValueError(f'metric "{self.name}" is not comparable (maximize=None)')

    @staticmethod
    def _to_float(other):
        if isinstance(other, Metric):
            return other.value
        else:
            return float(other)

    def __eq__(self, other):
        self._check_comparable()
        other = self._to_float(other)
        return abs(self.value - other) < self.comparison_tol

    def __gt__(self, other):
        self._check_comparable()
        other = self._to_float(other)
        if self.maximize is True:
            return self.value > other + self.comparison_tol
        else:
            return self.value < other - self.comparison_tol


class PercentageMetric(Metric):
    def __init__(
        self,
        name,
        precision: int = 2,
        show_percent: bool = True,
        maximize: bool = True,
        **kwargs,
    ):
        """
        precision: decimal places after the percentage point. E.g.
            precision=2 corresponds to 35.67% (show_percent=True)
                or 0.3567 (show_percent=False)
        """
        self.precision = int(precision)
        assert self.precision >= 1
        self.show_percent = show_percent

        super().__init__(
            name=name, maximize=maximize, formatter=self._format_acc, **kwargs,
        )

    def _format_acc(self, value):
        if self.show_percent:
            return f"{value * 100:.{self.precision}f}%"
        else:
            return f"{value:.{self.precision + 2}f}"

    def update(self, value: float, batch_size: int = 1):
        assert 0.0 <= value <= 1.0
        super().update(value, batch_size=batch_size)


class ScoreMetric(Metric):
    def __init__(
        self,
        name,
        precision: int = 4,
        scientific: bool = False,
        maximize: bool = True,
        **kwargs,
    ):
        self.precision = int(precision)
        assert self.precision >= 1
        self.scientific = scientific

        super().__init__(
            name=name, maximize=maximize, formatter=self._format_loss, **kwargs,
        )

    def _format_loss(self, value):
        return f"{value:.{self.precision}{'e' if self.scientific else 'g'}}"


class LossMetric(ScoreMetric):
    def __init__(self, name, **kwargs):
        super().__init__(name=name, maximize=False, **kwargs)


class TimingMetric(Metric):
    def __init__(
        self, name, unit: str = "auto", precision: int = 2, **kwargs,
    ):
        """
        precision: decimal places after the percentage point. E.g.
            precision=2 corresponds to 35.67% (show_percent=True)
                or 0.3567 (show_percent=False)
        """
        self.precision = int(precision)
        assert self.precision >= 1
        self.unit = unit

        super().__init__(
            name=name,
            maximize=None,
            formatter=self._format_time,
            sync_workers=False,
            **kwargs,
        )

    def _format_time(self, value):
        return U.Timer.pformat(value, unit=self.unit, float_fmt=f".{self.precision}f")

    def update(self, value: float, batch_size: int = 1):
        assert value >= 0.0
        super().update(value, batch_size=batch_size)
